// This file is made available under Elastic License 2.0.
// This file is based on code available under the Apache license here:
//   https://github.com/apache/incubator-doris/blob/master/fe/fe-core/src/main/java/org/apache/doris/transaction/PartitionCommitInfo.java

// Licensed to the Apache Software Foundation (ASF) under one
// or more contributor license agreements.  See the NOTICE file
// distributed with this work for additional information
// regarding copyright ownership.  The ASF licenses this file
// to you under the Apache License, Version 2.0 (the
// "License"); you may not use this file except in compliance
// with the License.  You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing,
// software distributed under the License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, either express or implied.  See the License for the
// specific language governing permissions and limitations
// under the License.

package com.starrocks.transaction;

import com.google.common.collect.Lists;
import com.google.gson.annotations.SerializedName;
import com.starrocks.catalog.Catalog;
import com.starrocks.common.FeMetaVersion;
import com.starrocks.common.io.Text;
import com.starrocks.common.io.Writable;
import com.starrocks.persist.gson.GsonUtils;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.util.List;

public class PartitionCommitInfo implements Writable {

    @SerializedName(value = "partitionId")
    private long partitionId;
    @SerializedName(value = "version")
    private long version;
    @SerializedName(value = "versionTime")
    private long versionTime;

    // For low cardinality string column with global dict
    // TODO(KKS): move invalidDictCacheColumns and validDictCacheColumns to TableCommitInfo
    // Currently, for support FE rollback, we persist the invalidDictCacheColumns in PartitionCommitInfo by json,
    // not TableCommitInfo.

    @SerializedName(value = "invalidColumns")
    private List<String> invalidDictCacheColumns = Lists.newArrayList();
    @SerializedName(value = "validColumns")
    private List<String> validDictCacheColumns = Lists.newArrayList();

    public PartitionCommitInfo() {

    }

    public PartitionCommitInfo(long partitionId, long version, long visibleTime) {
        super();
        this.partitionId = partitionId;
        this.version = version;
        this.versionTime = visibleTime;
    }

    public PartitionCommitInfo(long partitionId, long version, long visibleTime,
                               List<String> invalidDictCacheColumns,
                               List<String> validDictCacheColumns) {
        super();
        this.partitionId = partitionId;
        this.version = version;
        this.versionTime = visibleTime;
        this.invalidDictCacheColumns = invalidDictCacheColumns;
        this.validDictCacheColumns = validDictCacheColumns;
    }

    @Override
    public void write(DataOutput out) throws IOException {
        String json = GsonUtils.GSON.toJson(this);
        Text.writeString(out, json);
    }

    public static PartitionCommitInfo read(DataInput in) throws IOException {
        if (Catalog.getCurrentCatalogJournalVersion() < FeMetaVersion.VERSION_88) {
            long partitionId = in.readLong();
            long version = in.readLong();
            in.readLong();
            return new PartitionCommitInfo(partitionId, version, System.currentTimeMillis());
        } else {
            String json = Text.readString(in);
            return GsonUtils.GSON.fromJson(json, PartitionCommitInfo.class);
        }
    }

    public long getPartitionId() {
        return partitionId;
    }

    public long getVersion() {
        return version;
    }

    public long getVersionTime() {
        return versionTime;
    }

    public List<String> getInvalidDictCacheColumns() {
        return invalidDictCacheColumns;
    }

    public List<String> getValidDictCacheColumns() {
        return validDictCacheColumns;
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder("partitionid=");
        sb.append(partitionId);
        sb.append(", version=").append(version);
        sb.append(", versionHash=").append(0);
        sb.append(", versionTime=").append(versionTime);
        return sb.toString();
    }
}
